@@ -1,5 +1,7 @@
-#include <fitz.h>
-#include <mupdf.h>
+#include "fitz.h"
+#include "mupdf.h"
+#include "muxps.h"
+#include "mucbz.h"
 #include "pdfapp.h"
 
 #ifndef UNICODE
@@ -17,6 +19,8 @@
 #define WM_MOUSEWHEEL 0x020A
 #endif
 
+#define MIN(x,y) ((x) < (y) ? (x) : (y))
+
 #define ID_ABOUT	0x1000
 #define ID_DOCINFO	0x1001
 
@@ -26,16 +30,64 @@ static HDC hdc;
 static HBRUSH bgbrush;
 static HBRUSH shbrush;
 static BITMAPINFO *dibinf;
-static HCURSOR arrowcurs, handcurs, waitcurs;
+static HCURSOR arrowcurs, handcurs, waitcurs, caretcurs;
 static LRESULT CALLBACK frameproc(HWND, UINT, WPARAM, LPARAM);
 static LRESULT CALLBACK viewproc(HWND, UINT, WPARAM, LPARAM);
+static int timer_pending = 0;
 
 static int justcopied = 0;
 
 static pdfapp_t gapp;
 
-static wchar_t wbuf[1024];
-static char filename[1024];
+#define PATH_MAX (1024)
+
+static wchar_t wbuf[PATH_MAX];
+static char filename[PATH_MAX];
+
+/*
+ * Create registry keys to associate MuPDF with PDF and XPS files.
+ */
+
+#define OPEN_KEY(parent, name, ptr) \
+	RegCreateKeyExA(parent, name, 0, 0, 0, KEY_WRITE, 0, &ptr, 0)
+
+#define SET_KEY(parent, name, value) \
+	RegSetValueExA(parent, name, 0, REG_SZ, value, strlen(value) + 1)
+
+void install_app(char *argv0)
+{
+	char buf[512];
+	HKEY software, classes, mupdf, dotpdf, dotxps;
+	HKEY shell, open, command, supported_types;
+	HKEY pdf_progids, xps_progids;
+
+	OPEN_KEY(HKEY_CURRENT_USER, "Software", software);
+	OPEN_KEY(software, "Classes", classes);
+	OPEN_KEY(classes, ".pdf", dotpdf);
+	OPEN_KEY(dotpdf, "OpenWithProgids", pdf_progids);
+	OPEN_KEY(classes, ".xps", dotxps);
+	OPEN_KEY(dotxps, "OpenWithProgids", xps_progids);
+	OPEN_KEY(classes, "MuPDF", mupdf);
+	OPEN_KEY(mupdf, "SupportedTypes", supported_types);
+	OPEN_KEY(mupdf, "shell", shell);
+	OPEN_KEY(shell, "open", open);
+	OPEN_KEY(open, "command", command);
+
+	sprintf(buf, "\"%s\" \"%%1\"", argv0);
+
+	SET_KEY(open, "FriendlyAppName", "MuPDF");
+	SET_KEY(command, "", buf);
+	SET_KEY(supported_types, ".pdf", "");
+	SET_KEY(supported_types, ".xps", "");
+	SET_KEY(pdf_progids, "MuPDF", "");
+	SET_KEY(xps_progids, "MuPDF", "");
+
+	RegCloseKey(dotxps);
+	RegCloseKey(dotpdf);
+	RegCloseKey(mupdf);
+	RegCloseKey(classes);
+	RegCloseKey(software);
+}
 
 /*
  * Dialog boxes
@@ -46,37 +98,81 @@ void winwarn(pdfapp_t *app, char *msg)
 	MessageBoxA(hwndframe, msg, "MuPDF: Warning", MB_ICONWARNING);
 }
 
-void winerror(pdfapp_t *app, fz_error error)
+void winerror(pdfapp_t *app, char *msg)
 {
-	char msgbuf[160 * 30];
-	int i;
-
-	/* TODO: redirect stderr to a log file and display here */
-	fz_catch(error, "displaying error message to user");
-
-	fz_strlcpy(msgbuf, "An error has occurred.\n\n", sizeof msgbuf);
-	for (i = 0; i < fz_geterrorcount(); i++)
-	{
-		fz_strlcat(msgbuf, fz_geterrorline(i), sizeof msgbuf);
-		fz_strlcat(msgbuf, "\n", sizeof msgbuf);
-	}
-
-	MessageBoxA(hwndframe, msgbuf, "MuPDF: Error", MB_ICONERROR);
+	MessageBoxA(hwndframe, msg, "MuPDF: Error", MB_ICONERROR);
 	exit(1);
 }
 
-void win32error(char *msg)
+void winalert(pdfapp_t *app, fz_alert_event *alert)
 {
-	LPSTR buf;
-	int code = GetLastError();
-	FormatMessageA(FORMAT_MESSAGE_ALLOCATE_BUFFER |
-		FORMAT_MESSAGE_FROM_SYSTEM |
-		FORMAT_MESSAGE_IGNORE_INSERTS,
-		NULL,
-		code,
-		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
-		(LPSTR)&buf, 0, NULL);
-	winerror(&gapp, fz_throw("%s:\n%s", msg, buf));
+	int buttons = MB_OK;
+	int icon = MB_ICONWARNING;
+	int pressed = FZ_ALERT_BUTTON_NONE;
+
+	switch (alert->icon_type)
+	{
+	case FZ_ALERT_ICON_ERROR:
+		icon = MB_ICONERROR;
+		break;
+	case FZ_ALERT_ICON_WARNING:
+		icon = MB_ICONWARNING;
+		break;
+	case FZ_ALERT_ICON_QUESTION:
+		icon = MB_ICONQUESTION;
+		break;
+	case FZ_ALERT_ICON_STATUS:
+		icon = MB_ICONINFORMATION;
+		break;
+	}
+
+	switch (alert->button_group_type)
+	{
+	case FZ_ALERT_BUTTON_GROUP_OK:
+		buttons = MB_OK;
+		break;
+	case FZ_ALERT_BUTTON_GROUP_OK_CANCEL:
+		buttons = MB_OKCANCEL;
+		break;
+	case FZ_ALERT_BUTTON_GROUP_YES_NO:
+		buttons = MB_YESNO;
+		break;
+	case FZ_ALERT_BUTTON_GROUP_YES_NO_CANCEL:
+		buttons = MB_YESNOCANCEL;
+		break;
+	}
+
+	pressed = MessageBoxA(hwndframe, alert->message, alert->title, icon|buttons);
+
+	switch (pressed)
+	{
+	case IDOK:
+		alert->button_pressed = FZ_ALERT_BUTTON_OK;
+		break;
+	case IDCANCEL:
+		alert->button_pressed = FZ_ALERT_BUTTON_CANCEL;
+		break;
+	case IDNO:
+		alert->button_pressed = FZ_ALERT_BUTTON_NO;
+		break;
+	case IDYES:
+		alert->button_pressed = FZ_ALERT_BUTTON_YES;
+	}
+}
+
+void winprint(pdfapp_t *app)
+{
+	MessageBoxA(hwndframe, "The MuPDF library supports printing, but this application currently does not", "Print document", MB_ICONWARNING);
+}
+
+int winsavequery(pdfapp_t *app)
+{
+	switch(MessageBoxA(hwndframe, "File has unsaved changes. Do you want to save", "MuPDF", MB_YESNOCANCEL))
+	{
+	case IDYES: return SAVE;
+	case IDNO: return DISCARD;
+	default: return CANCEL;
+	}
 }
 
 int winfilename(wchar_t *buf, int len)
@@ -90,13 +186,80 @@ int winfilename(wchar_t *buf, int len)
 	ofn.nMaxFile = len;
 	ofn.lpstrInitialDir = NULL;
 	ofn.lpstrTitle = L"MuPDF: Open PDF file";
-	ofn.lpstrFilter = L"PDF Files (*.pdf)\0*.pdf\0All Files\0*\0\0";
+	ofn.lpstrFilter = L"Documents (*.pdf;*.xps;*.cbz;*.zip)\0*.zip;*.cbz;*.xps;*.pdf\0PDF Files (*.pdf)\0*.pdf\0XPS Files (*.xps)\0*.xps\0CBZ Files (*.cbz;*.zip)\0*.zip;*.cbz\0All Files\0*\0\0";
 	ofn.Flags = OFN_FILEMUSTEXIST|OFN_HIDEREADONLY;
 	return GetOpenFileNameW(&ofn);
 }
 
+int wingetsavepath(pdfapp_t *app, char *buf, int len)
+{
+	wchar_t twbuf[PATH_MAX];
+	OPENFILENAME ofn;
+
+	wcscpy(twbuf, wbuf);
+	memset(&ofn, 0, sizeof(OPENFILENAME));
+	ofn.lStructSize = sizeof(OPENFILENAME);
+	ofn.hwndOwner = hwndframe;
+	ofn.lpstrFile = twbuf;
+	ofn.nMaxFile = PATH_MAX;
+	ofn.lpstrInitialDir = NULL;
+	ofn.lpstrTitle = L"MuPDF: Save PDF file";
+	ofn.lpstrFilter = L"Documents (*.pdf;*.xps;*.cbz;*.zip)\0*.zip;*.cbz;*.xps;*.pdf\0PDF Files (*.pdf)\0*.pdf\0XPS Files (*.xps)\0*.xps\0CBZ Files (*.cbz;*.zip)\0*.zip;*.cbz\0All Files\0*\0\0";
+	ofn.Flags = OFN_HIDEREADONLY;
+	if (GetSaveFileName(&ofn))
+	{
+		int code = WideCharToMultiByte(CP_UTF8, 0, twbuf, -1, buf, MIN(PATH_MAX, len), NULL, NULL);
+		if (code == 0)
+		{
+			winerror(&gapp, "cannot convert filename to utf-8");
+			return 0;
+		}
+
+		wcscpy(wbuf, twbuf);
+		strcpy(filename, buf);
+		return 1;
+	}
+	else
+	{
+		return 0;
+	}
+}
+
+void winreplacefile(char *source, char *target)
+{
+	wchar_t wsource[PATH_MAX];
+	wchar_t wtarget[PATH_MAX];
+
+	int sz = MultiByteToWideChar(CP_UTF8, 0, source, -1, wsource, PATH_MAX);
+	if (sz == 0)
+	{
+		winerror(&gapp, "cannot convert filename to Unicode");
+		return;
+	}
+
+	sz = MultiByteToWideChar(CP_UTF8, 0, target, -1, wtarget, PATH_MAX);
+	if (sz == 0)
+	{
+		winerror(&gapp, "cannot convert filename to Unicode");
+		return;
+	}
+
+#if (_WIN32_WINNT >= 0x0500)
+	ReplaceFile(wtarget, wsource, NULL, REPLACEFILE_IGNORE_MERGE_ERRORS, NULL, NULL);
+#else
+	DeleteFile(wtarget);
+	MoveFile(wsource, wtarget);
+#endif
+}
+
 static char pd_filename[256] = "The file is encrypted.";
 static char pd_password[256] = "";
+static char td_textinput[1024] = "";
+static int td_retry = 0;
+static int cd_nopts;
+static int *cd_nvals;
+static char **cd_opts;
+static char **cd_vals;
 static int pd_okay = 0;
 
 INT CALLBACK
@@ -125,9 +288,94 @@ dlogpassproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 	return FALSE;
 }
 
+INT CALLBACK
+dlogtextproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
+{
+	switch(message)
+	{
+	case WM_INITDIALOG:
+		SetDlgItemTextA(hwnd, 3, td_textinput);
+		if (!td_retry)
+			ShowWindow(GetDlgItem(hwnd, 4), SW_HIDE);
+		return TRUE;
+	case WM_COMMAND:
+		switch(wParam)
+		{
+		case 1:
+			pd_okay = 1;
+			GetDlgItemTextA(hwnd, 3, td_textinput, sizeof td_textinput);
+			EndDialog(hwnd, 1);
+			return TRUE;
+		case 2:
+			pd_okay = 0;
+			EndDialog(hwnd, 1);
+			return TRUE;
+		}
+		break;
+	case WM_CTLCOLORSTATIC:
+		if ((HWND)lParam == GetDlgItem(hwnd, 4))
+		{
+			SetTextColor((HDC)wParam, RGB(255,0,0));
+			SetBkMode((HDC)wParam, TRANSPARENT);
+
+			return (INT)GetStockObject(NULL_BRUSH);
+		}
+		break;
+	}
+	return FALSE;
+}
+
+INT CALLBACK
+dlogchoiceproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
+{
+	HWND listbox;
+	int i;
+	int item;
+	int sel;
+	switch(message)
+	{
+	case WM_INITDIALOG:
+		listbox = GetDlgItem(hwnd, 3);
+		for (i = 0; i < cd_nopts; i++)
+			SendMessageA(listbox, LB_ADDSTRING, 0, (LPARAM)cd_opts[i]);
+
+		/* FIXME: handle multiple select */
+		if (*cd_nvals > 0)
+		{
+			item = SendMessageA(listbox, LB_FINDSTRINGEXACT, -1, (LPARAM)cd_vals[0]);
+			if (item != LB_ERR)
+				SendMessageA(listbox, LB_SETCURSEL, item, 0);
+		}
+		return TRUE;
+	case WM_COMMAND:
+		switch(wParam)
+		{
+		case 1:
+			listbox = GetDlgItem(hwnd, 3);
+			*cd_nvals = 0;
+			for (i = 0; i < cd_nopts; i++)
+			{
+				item = SendMessageA(listbox, LB_FINDSTRINGEXACT, -1, (LPARAM)cd_opts[i]);
+				sel = SendMessageA(listbox, LB_GETSEL, item, 0);
+				if (sel && sel != LB_ERR)
+					cd_vals[(*cd_nvals)++] = cd_opts[i];
+			}
+			pd_okay = 1;
+			EndDialog(hwnd, 1);
+			return TRUE;
+		case 2:
+			pd_okay = 0;
+			EndDialog(hwnd, 1);
+			return TRUE;
+		}
+		break;
+	}
+	return FALSE;
+}
+
 char *winpassword(pdfapp_t *app, char *filename)
 {
-	char buf[124], *s;
+	char buf[1024], *s;
 	int code;
 	strcpy(buf, filename);
 	s = buf;
@@ -138,18 +386,43 @@ char *winpassword(pdfapp_t *app, char *filename)
 	sprintf(pd_filename, "The file \"%s\" is encrypted.", s);
 	code = DialogBoxW(NULL, L"IDD_DLOGPASS", hwndframe, dlogpassproc);
 	if (code <= 0)
-		win32error("cannot create password dialog");
+		winerror(app, "cannot create password dialog");
 	if (pd_okay)
 		return pd_password;
-	return nil;
+	return NULL;
+}
+
+char *wintextinput(pdfapp_t *app, char *inittext, int retry)
+{
+	int code;
+	td_retry = retry;
+	fz_strlcpy(td_textinput, inittext ? inittext : "", sizeof td_textinput);
+	code = DialogBoxW(NULL, L"IDD_DLOGTEXT", hwndframe, dlogtextproc);
+	if (code <= 0)
+		winerror(app, "cannot create text input dialog");
+	if (pd_okay)
+		return td_textinput;
+	return NULL;
+}
+
+int winchoiceinput(pdfapp_t *app, int nopts, char *opts[], int *nvals, char *vals[])
+{
+	int code;
+	cd_nopts = nopts;
+	cd_nvals = nvals;
+	cd_opts = opts;
+	cd_vals = vals;
+	code = DialogBoxW(NULL, L"IDD_DLOGLIST", hwndframe, dlogchoiceproc);
+	if (code <= 0)
+		winerror(app, "cannot create text input dialog");
+	return pd_okay;
 }
 
 INT CALLBACK
 dloginfoproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 {
 	char buf[256];
-	pdf_xref *xref = gapp.xref;
-	fz_obj *info, *obj;
+	fz_document *doc = gapp.doc;
 
 	switch(message)
 	{
@@ -157,63 +430,55 @@ dloginfoproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 
 		SetDlgItemTextW(hwnd, 0x10, wbuf);
 
-		sprintf(buf, "PDF %d.%d", xref->version / 10, xref->version % 10);
+		if (fz_meta(doc, FZ_META_FORMAT_INFO, buf, 256) < 0)
+		{
+			SetDlgItemTextA(hwnd, 0x11, "Unknown");
+			SetDlgItemTextA(hwnd, 0x12, "None");
+			SetDlgItemTextA(hwnd, 0x13, "n/a");
+			return TRUE;
+		}
+
 		SetDlgItemTextA(hwnd, 0x11, buf);
 
-		if (xref->crypt)
+		if (fz_meta(doc, FZ_META_CRYPT_INFO, buf, 256) == 0)
 		{
-			sprintf(buf, "Standard %d bit %s", xref->crypt->length,
-				xref->crypt->strf.method == PDF_CRYPT_AESV2 ? "AES" : "RC4");
 			SetDlgItemTextA(hwnd, 0x12, buf);
-			strcpy(buf, "");
-			if (xref->crypt->p & (1 << 2))
-				strcat(buf, "print, ");
-			if (xref->crypt->p & (1 << 3))
-				strcat(buf, "modify, ");
-			if (xref->crypt->p & (1 << 4))
-				strcat(buf, "copy, ");
-			if (xref->crypt->p & (1 << 5))
-				strcat(buf, "annotate, ");
-			if (strlen(buf) > 2)
-				buf[strlen(buf)-2] = 0;
-			else
-				strcpy(buf, "none");
-			SetDlgItemTextA(hwnd, 0x13, buf);
 		}
 		else
 		{
 			SetDlgItemTextA(hwnd, 0x12, "None");
-			SetDlgItemTextA(hwnd, 0x13, "n/a");
 		}
+		buf[0] = 0;
+		if (fz_meta(doc, FZ_META_HAS_PERMISSION, NULL, FZ_PERMISSION_PRINT) == 0)
+			strcat(buf, "print, ");
+		if (fz_meta(doc, FZ_META_HAS_PERMISSION, NULL, FZ_PERMISSION_CHANGE) == 0)
+			strcat(buf, "modify, ");
+		if (fz_meta(doc, FZ_META_HAS_PERMISSION, NULL, FZ_PERMISSION_COPY) == 0)
+			strcat(buf, "copy, ");
+		if (fz_meta(doc, FZ_META_HAS_PERMISSION, NULL, FZ_PERMISSION_NOTES) == 0)
+			strcat(buf, "annotate, ");
+		if (strlen(buf) > 2)
+			buf[strlen(buf)-2] = 0;
+		else
+			strcpy(buf, "None");
+		SetDlgItemTextA(hwnd, 0x13, buf);
 
-		info = fz_dictgets(xref->trailer, "Info");
-		if (!info)
-			return TRUE;
-
-#define SETUCS(ID) \
+#define SETUTF8(ID, STRING) \
 		{ \
-			unsigned short *ucs; \
-			ucs = pdf_toucs2(obj); \
-			SetDlgItemTextW(hwnd, ID, ucs); \
-			fz_free(ucs); \
+			*(char **)buf = STRING; \
+			if (fz_meta(doc, FZ_META_INFO, buf, 256) <= 0) \
+				buf[0] = 0; \
+			SetDlgItemTextA(hwnd, ID, buf); \
 		}
 
-		if ((obj = fz_dictgets(info, "Title")))
-			SETUCS(0x20);
-		if ((obj = fz_dictgets(info, "Author")))
-			SETUCS(0x21);
-		if ((obj = fz_dictgets(info, "Subject")))
-			SETUCS(0x22);
-		if ((obj = fz_dictgets(info, "Keywords")))
-			SETUCS(0x23);
-		if ((obj = fz_dictgets(info, "Creator")))
-			SETUCS(0x24);
-		if ((obj = fz_dictgets(info, "Producer")))
-			SETUCS(0x25);
-		if ((obj = fz_dictgets(info, "CreationDate")))
-			SETUCS(0x26);
-		if ((obj = fz_dictgets(info, "ModDate")))
-			SETUCS(0x27);
+		SETUTF8(0x20, "Title");
+		SETUTF8(0x21, "Author");
+		SETUTF8(0x22, "Subject");
+		SETUTF8(0x23, "Keywords");
+		SETUTF8(0x24, "Creator");
+		SETUTF8(0x25, "Producer");
+		SETUTF8(0x26, "CreationDate");
+		SETUTF8(0x27, "ModDate");
 		return TRUE;
 
 	case WM_COMMAND:
@@ -227,7 +492,7 @@ void info()
 {
 	int code = DialogBoxW(NULL, L"IDD_DLOGINFO", hwndframe, dloginfoproc);
 	if (code <= 0)
-		win32error("cannot create info dialog");
+		winerror(&gapp, "cannot create info dialog");
 }
 
 INT CALLBACK
@@ -250,7 +515,7 @@ void winhelp(pdfapp_t*app)
 {
 	int code = DialogBoxW(NULL, L"IDD_DLOGABOUT", hwndframe, dlogaboutproc);
 	if (code <= 0)
-		win32error("cannot create help dialog");
+		winerror(&gapp, "cannot create help dialog");
 }
 
 /*
@@ -278,7 +543,7 @@ void winopen()
 	wc.lpszClassName = L"FrameWindow";
 	a = RegisterClassW(&wc);
 	if (!a)
-		win32error("cannot register frame window class");
+		winerror(&gapp, "cannot register frame window class");
 
 	/* Create and register window view class */
 	memset(&wc, 0, sizeof(wc));
@@ -294,7 +559,7 @@ void winopen()
 	wc.lpszClassName = L"ViewWindow";
 	a = RegisterClassW(&wc);
 	if (!a)
-		win32error("cannot register view window class");
+		winerror(&gapp, "cannot register view window class");
 
 	/* Get screen size */
 	SystemParametersInfo(SPI_GETWORKAREA, 0, &r, 0);
@@ -305,6 +570,7 @@ void winopen()
 	arrowcurs = LoadCursor(NULL, IDC_ARROW);
 	handcurs = LoadCursor(NULL, IDC_HAND);
 	waitcurs = LoadCursor(NULL, IDC_WAIT);
+	caretcurs = LoadCursor(NULL, IDC_IBEAM);
 
 	/* And a background color */
 	bgbrush = CreateSolidBrush(RGB(0x70,0x70,0x70));
@@ -312,7 +578,7 @@ void winopen()
 
 	/* Init DIB info for buffer */
 	dibinf = malloc(sizeof(BITMAPINFO) + 12);
-	assert(dibinf != NULL);
+	assert(dibinf);
 	dibinf->bmiHeader.biSize = sizeof(dibinf->bmiHeader);
 	dibinf->bmiHeader.biPlanes = 1;
 	dibinf->bmiHeader.biBitCount = 32;
@@ -335,7 +601,7 @@ void winopen()
 	0, // program instance handle
 	0); // creation parameters
 	if (!hwndframe)
-		win32error("cannot create frame: %s");
+		winerror(&gapp, "cannot create frame");
 
 	hwndview = CreateWindowW(L"ViewWindow", // window class name
 	NULL,
@@ -344,7 +610,7 @@ void winopen()
 	CW_USEDEFAULT, CW_USEDEFAULT,
 	hwndframe, 0, 0, 0);
 	if (!hwndview)
-		win32error("cannot create view: %s");
+		winerror(&gapp, "cannot create view");
 
 	hdc = NULL;
 
@@ -360,8 +626,11 @@ void winopen()
 
 void winclose(pdfapp_t *app)
 {
-	pdfapp_close(app);
-	exit(0);
+	if (pdfapp_preclose(app))
+	{
+		pdfapp_close(app);
+		exit(0);
+	}
 }
 
 void wincursor(pdfapp_t *app, int curs)
@@ -372,6 +641,8 @@ void wincursor(pdfapp_t *app, int curs)
 		SetCursor(handcurs);
 	if (curs == WAIT)
 		SetCursor(waitcurs);
+	if (curs == CARET)
+		SetCursor(caretcurs);
 }
 
 void wintitle(pdfapp_t *app, char *title)
@@ -384,7 +655,7 @@ void wintitle(pdfapp_t *app, char *title)
 	sp = title;
 	while (*sp && dp < wide + 255)
 	{
-		sp += chartorune(&rune, sp);
+		sp += fz_chartorune(&rune, sp);
 		*dp++ = rune;
 	}
 	*dp = 0;
@@ -409,33 +680,48 @@ void windrawstring(pdfapp_t *app, int x, int y, char *s)
 	TextOutA(hdc, x, y - 12, s, strlen(s));
 }
 
+void winblitsearch()
+{
+	if (gapp.isediting)
+	{
+		char buf[sizeof(gapp.search) + 50];
+		sprintf(buf, "Search: %s", gapp.search);
+		windrawrect(&gapp, 0, 0, gapp.winw, 30);
+		windrawstring(&gapp, 10, 20, buf);
+	}
+}
+
 void winblit()
 {
+	int image_w = fz_pixmap_width(gapp.ctx, gapp.image);
+	int image_h = fz_pixmap_height(gapp.ctx, gapp.image);
+	int image_n = fz_pixmap_components(gapp.ctx, gapp.image);
+	unsigned char *samples = fz_pixmap_samples(gapp.ctx, gapp.image);
 	int x0 = gapp.panx;
 	int y0 = gapp.pany;
-	int x1 = gapp.panx + gapp.image->w;
-	int y1 = gapp.pany + gapp.image->h;
+	int x1 = gapp.panx + image_w;
+	int y1 = gapp.pany + image_h;
 	RECT r;
 
 	if (gapp.image)
 	{
 		if (gapp.iscopying || justcopied)
 		{
-			pdfapp_invert(&gapp, gapp.selr);
+			pdfapp_invert(&gapp, &gapp.selr);
 			justcopied = 1;
 		}
 
 		pdfapp_inverthit(&gapp);
 
-		dibinf->bmiHeader.biWidth = gapp.image->w;
-		dibinf->bmiHeader.biHeight = -gapp.image->h;
-		dibinf->bmiHeader.biSizeImage = gapp.image->h * 4;
+		dibinf->bmiHeader.biWidth = image_w;
+		dibinf->bmiHeader.biHeight = -image_h;
+		dibinf->bmiHeader.biSizeImage = image_h * 4;
 
-		if (gapp.image->n == 2)
+		if (image_n == 2)
 		{
-			int i = gapp.image->w * gapp.image->h;
+			int i = image_w * image_h;
 			unsigned char *color = malloc(i*4);
-			unsigned char *s = gapp.image->samples;
+			unsigned char *s = samples;
 			unsigned char *d = color;
 			for (; i > 0 ; i--)
 			{
@@ -444,16 +730,16 @@ void winblit()
 				d += 4;
 			}
 			SetDIBitsToDevice(hdc,
-				gapp.panx, gapp.pany, gapp.image->w, gapp.image->h,
-				0, 0, 0, gapp.image->h, color,
+				gapp.panx, gapp.pany, image_w, image_h,
+				0, 0, 0, image_h, color,
 				dibinf, DIB_RGB_COLORS);
 			free(color);
 		}
-		if (gapp.image->n == 4)
+		if (image_n == 4)
 		{
 			SetDIBitsToDevice(hdc,
-				gapp.panx, gapp.pany, gapp.image->w, gapp.image->h,
-				0, 0, 0, gapp.image->h, gapp.image->samples,
+				gapp.panx, gapp.pany, image_w, image_h,
+				0, 0, 0, image_h, samples,
 				dibinf, DIB_RGB_COLORS);
 		}
 
@@ -461,7 +747,7 @@ void winblit()
 
 		if (gapp.iscopying || justcopied)
 		{
-			pdfapp_invert(&gapp, gapp.selr);
+			pdfapp_invert(&gapp, &gapp.selr);
 			justcopied = 1;
 		}
 	}
@@ -490,13 +776,7 @@ void winblit()
 	r.bottom = y1;
 	FillRect(hdc, &r, shbrush);
 
-	if (gapp.isediting)
-	{
-		char buf[sizeof(gapp.search) + 50];
-		sprintf(buf, "Search: %s", gapp.search);
-		windrawrect(&gapp, 0, 0, gapp.winw, 30);
-		windrawstring(&gapp, 10, 20, buf);
-	}
+	winblitsearch();
 }
 
 void winresize(pdfapp_t *app, int w, int h)
@@ -513,6 +793,34 @@ void winrepaint(pdfapp_t *app)
 	InvalidateRect(hwndview, NULL, 0);
 }
 
+void winrepaintsearch(pdfapp_t *app)
+{
+	// TODO: invalidate only search area and
+	// call only search redraw routine.
+	InvalidateRect(hwndview, NULL, 0);
+}
+
+void winfullscreen(pdfapp_t *app, int state)
+{
+	static WINDOWPLACEMENT savedplace;
+	static int isfullscreen = 0;
+	if (state && !isfullscreen)
+	{
+		GetWindowPlacement(hwndframe, &savedplace);
+		SetWindowLong(hwndframe, GWL_STYLE, WS_POPUP | WS_VISIBLE);
+		SetWindowPos(hwndframe, NULL, 0, 0, 0, 0, SWP_NOSIZE | SWP_NOMOVE | SWP_NOZORDER | SWP_FRAMECHANGED);
+		ShowWindow(hwndframe, SW_SHOWMAXIMIZED);
+		isfullscreen = 1;
+	}
+	if (!state && isfullscreen)
+	{
+		SetWindowLong(hwndframe, GWL_STYLE, WS_OVERLAPPEDWINDOW);
+		SetWindowPos(hwndframe, NULL, 0, 0, 0, 0, SWP_NOSIZE | SWP_NOMOVE | SWP_NOZORDER | SWP_FRAMECHANGED);
+		SetWindowPlacement(hwndframe, &savedplace);
+		isfullscreen = 0;
+	}
+}
+
 /*
  * Event handling
  */
@@ -545,15 +853,8 @@ void windocopy(pdfapp_t *app)
 
 void winreloadfile(pdfapp_t *app)
 {
-	int fd;
-
 	pdfapp_close(app);
-
-	fd = _wopen(wbuf, O_BINARY | O_RDONLY, 0666);
-	if (fd < 0)
-		winerror(&gapp, fz_throw("cannot reload file '%s'", filename));
-
-	pdfapp_open(app, filename, fd);
+	pdfapp_open(app, filename, 1);
 }
 
 void winopenuri(pdfapp_t *app, char *buf)
@@ -561,8 +862,24 @@ void winopenuri(pdfapp_t *app, char *buf)
 	ShellExecuteA(hwndframe, "open", buf, 0, 0, SW_SHOWNORMAL);
 }
 
+#define OUR_TIMER_ID 1
+
+void winadvancetimer(pdfapp_t *app, float delay)
+{
+	timer_pending = 1;
+	SetTimer(hwndview, OUR_TIMER_ID, (unsigned int)(1000*delay), NULL);
+}
+
+static void killtimer(pdfapp_t *app)
+{
+	timer_pending = 0;
+}
+
 void handlekey(int c)
 {
+	if (timer_pending)
+		killtimer(&gapp);
+
 	if (GetCapture() == hwndview)
 		return;
 
@@ -572,7 +889,7 @@ void handlekey(int c)
 		winrepaint(&gapp);
 	}
 
-	/* translate VK into ascii equivalents */
+	/* translate VK into ASCII equivalents */
 	if (c > 256)
 	{
 		switch (c - 256)
@@ -594,6 +911,9 @@ void handlekey(int c)
 
 void handlemouse(int x, int y, int btn, int state)
 {
+	if (state != 0 && timer_pending)
+		killtimer(&gapp);
+
 	if (state != 0 && justcopied)
 	{
 		justcopied = 0;
@@ -635,6 +955,8 @@ frameproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 			info();
 			return 0;
 		}
+		if (wParam == SC_MAXIMIZE)
+			gapp.shrinkwrap = 0;
 		break;
 
 	case WM_SIZE:
@@ -655,6 +977,10 @@ frameproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 	case WM_NOTIFY:
 	case WM_COMMAND:
 		return SendMessage(hwndview, message, wParam, lParam);
+
+	case WM_CLOSE:
+		if (!pdfapp_preclose(&gapp))
+			return 0;
 	}
 
 	return DefWindowProc(hwnd, message, wParam, lParam);
@@ -689,6 +1015,7 @@ viewproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 		winblit();
 		hdc = NULL;
 		EndPaint(hwnd, &ps);
+		pdfapp_postblit(&gapp);
 		return 0;
 	}
 
@@ -740,6 +1067,17 @@ viewproc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 			handlekey(LOWORD(wParam) & MK_SHIFT ? '-' : 'j');
 		return 0;
 
+	/* Timer */
+	case WM_TIMER:
+		if (wParam == OUR_TIMER_ID && timer_pending && gapp.presentation_mode)
+		{
+			timer_pending = 0;
+			handlekey(VK_RIGHT + 256);
+			handlemouse(oldx, oldy, 0, 0); /* update cursor */
+			return 0;
+		}
+		break;
+
 	/* Keyboard events */
 
 	case WM_KEYDOWN:
@@ -781,15 +1119,22 @@ WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShow
 {
 	int argc;
 	LPWSTR *argv = CommandLineToArgvW(GetCommandLineW(), &argc);
+	char argv0[256];
 	MSG msg;
-	int fd;
 	int code;
+	fz_context *ctx;
 
-	fz_accelerate();
+	ctx = fz_new_context(NULL, NULL, FZ_STORE_DEFAULT);
+	if (!ctx)
+	{
+		fprintf(stderr, "cannot initialise context\n");
+		exit(1);
+	}
+	pdfapp_init(ctx, &gapp);
 
-	pdfapp_init(&gapp);
+	GetModuleFileNameA(NULL, argv0, sizeof argv0);
+	install_app(argv0);
 
-	/* associate(argv[0]); */
 	winopen();
 
 	if (argc == 2)
@@ -802,15 +1147,11 @@ WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShow
 			exit(0);
 	}
 
-	fd = _wopen(wbuf, O_BINARY | O_RDONLY, 0666);
-	if (fd < 0)
-		winerror(&gapp, fz_throw("cannot open file '%s'", filename));
-
 	code = WideCharToMultiByte(CP_UTF8, 0, wbuf, -1, filename, sizeof filename, NULL, NULL);
 	if (code == 0)
-		win32error("cannot convert filename to utf-8");
+		winerror(&gapp, "cannot convert filename to utf-8");
 
-	pdfapp_open(&gapp, filename, fd);
+	pdfapp_open(&gapp, filename, 0);
 
 	while (GetMessage(&msg, NULL, 0, 0))
 	{
