@@ -1,4 +1,4 @@
-#include "fitz.h"
+#include "fitz-internal.h"
 
 /* Fax G3/G4 decoder */
 
@@ -31,7 +31,7 @@ enum
 	cfd_white_initial_bits = 8,
 	cfd_black_initial_bits = 7,
 	cfd_2d_initial_bits = 7,
-	cfd_uncompressed_initial_bits = 6	/* must be 6 */
+	cfd_uncompressed_initial_bits = 6 /* must be 6 */
 };
 
 /* non-run codes in tables */
@@ -57,7 +57,7 @@ enum
 };
 
 /* White decoding table. */
-const cfd_node cf_white_decode[] = {
+static const cfd_node cf_white_decode[] = {
 	{256,12},{272,12},{29,8},{30,8},{45,8},{46,8},{22,7},{22,7},
 	{23,7},{23,7},{47,8},{48,8},{13,6},{13,6},{13,6},{13,6},{20,7},
 	{20,7},{33,8},{34,8},{35,8},{36,8},{37,8},{38,8},{19,7},{19,7},
@@ -95,7 +95,7 @@ const cfd_node cf_white_decode[] = {
 };
 
 /* Black decoding table. */
-const cfd_node cf_black_decode[] = {
+static const cfd_node cf_black_decode[] = {
 	{128,12},{160,13},{224,12},{256,12},{10,7},{11,7},{288,12},{12,7},
 	{9,6},{9,6},{8,6},{8,6},{7,5},{7,5},{7,5},{7,5},{6,4},{6,4},{6,4},
 	{6,4},{6,4},{6,4},{6,4},{6,4},{5,4},{5,4},{5,4},{5,4},{5,4},{5,4},
@@ -134,7 +134,7 @@ const cfd_node cf_black_decode[] = {
 };
 
 /* 2-D decoding table. */
-const cfd_node cf_2d_decode[] = {
+static const cfd_node cf_2d_decode[] = {
 	{128,11},{144,10},{6,7},{0,7},{5,6},{5,6},{1,6},{1,6},{-4,4},
 	{-4,4},{-4,4},{-4,4},{-4,4},{-4,4},{-4,4},{-4,4},{-5,3},{-5,3},
 	{-5,3},{-5,3},{-5,3},{-5,3},{-5,3},{-5,3},{-5,3},{-5,3},{-5,3},
@@ -152,8 +152,8 @@ const cfd_node cf_2d_decode[] = {
 	{-1,0},{-1,0},{-1,0},{-1,0},{-1,0},{-1,0},{-1,0},{-1,0},{-3,3}
 };
 
-/* Uncompresssed decoding table. */
-const cfd_node cf_uncompressed_decode[] = {
+/* Uncompressed decoding table. */
+static const cfd_node cf_uncompressed_decode[] = {
 	{64,12},{5,6},{4,5},{4,5},{3,4},{3,4},{3,4},{3,4},{2,3},{2,3},
 	{2,3},{2,3},{2,3},{2,3},{2,3},{2,3},{1,2},{1,2},{1,2},{1,2},{1,2},
 	{1,2},{1,2},{1,2},{1,2},{1,2},{1,2},{1,2},{1,2},{1,2},{1,2},{1,2},
@@ -170,72 +170,103 @@ const cfd_node cf_uncompressed_decode[] = {
 
 /* bit magic */
 
-static inline void
-printbits(FILE *f, int code, int nbits)
-{
-	int n, b;
-	for (n = nbits - 1; n >= 0; n--)
-	{
-		b = (code >> n) & 1;
-		fprintf(f, "%c", b ? '1' : '0');
-	}
-}
-
-static inline int
-getbit(const unsigned char *buf, int x)
+static inline int getbit(const unsigned char *buf, int x)
 {
 	return ( buf[x >> 3] >> ( 7 - (x & 7) ) ) & 1;
 }
 
-static inline void
-printline(FILE *f, unsigned char *line, int w)
-{
-	int i;
-	for (i = 0; i < w; i++)
-		fprintf(f, "%c", getbit(line, i) ? '#' : '.');
-	fprintf(f, "\n");
-}
+static const unsigned char mask[8] = {
+	0x7F, 0x3F, 0x1F, 0x0F, 0x07, 0x03, 0x01, 0
+};
+
+static const unsigned char clz[256] = {
+	8, 7, 6, 6, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4,
+	3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
+	2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
+	2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
+	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
+	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
+	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
+	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
+};
 
 static inline int
-findchanging(const unsigned char *line, int x, int w)
+find_changing(const unsigned char *line, int x, int w)
 {
-	int a, b;
+	int a, b, m, W;
 
 	if (!line)
 		return w;
 
-	if (x == -1)
+	/* We assume w > 0, -1 <= x < w */
+	if (x < 0)
 	{
-		a = 0;
 		x = 0;
+		m = 0xFF;
 	}
 	else
 	{
-		a = getbit(line, x);
-		x++;
+		/* Mask out the bits we've already used (including the one
+		 * we started from) */
+		m = mask[x & 7];
 	}
-
-	while (x < w)
+	/* We have 'w' pixels (bits) in line. The last pixel that can be
+	 * safely accessed is the (w-1)th bit of line.
+	 * By taking W = w>>3, we know that the first W bytes of line are
+	 * full, with w&7 stray bits following. */
+	W = w>>3;
+	x >>= 3;
+	a = line[x]; /* Safe as x < w => x <= w-1 => x>>3 <= (w-1)>>3 */
+	b = a ^ (a>>1);
+	b &= m;
+	if (x >= W)
 	{
-		b = getbit(line, x);
-		if (a != b)
-			break;
-		x++;
+		/* Within the last byte already */
+		x = (x<<3) + clz[b];
+		if (x > w)
+			x = w;
+		return x;
 	}
-
+	while (b == 0)
+	{
+		if (++x >= W)
+			goto nearend;
+		b = a & 1;
+		a = line[x];
+		b = (b<<7) ^ a ^ (a>>1);
+	}
+	return (x<<3) + clz[b];
+nearend:
+	/* We have less than a byte to go. If no stray bits, exit now. */
+	if ((x<<3) == w)
+		return w;
+	b = a&1;
+	a = line[x];
+	b = (b<<7) ^ a ^ (a>>1);
+	x = (x<<3) + clz[b];
+	if (x > w)
+		x = w;
 	return x;
 }
 
 static inline int
-findchangingcolor(const unsigned char *line, int x, int w, int color)
+find_changing_color(const unsigned char *line, int x, int w, int color)
 {
-	if (!line)
+	if (!line || x >= w)
 		return w;
 
-	x = findchanging(line, x, w);
+	x = find_changing(line, (x > 0 || !color) ? x : -1, w);
 
 	if (x < w && getbit(line, x) != color)
-		x = findchanging(line, x, w);
+		x = find_changing(line, x, w);
 
 	return x;
 }
@@ -248,11 +279,13 @@ static const unsigned char rm[8] = {
 	0x00, 0x80, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC, 0xFE
 };
 
-static inline void
-setbits(unsigned char *line, int x0, int x1)
+static inline void setbits(unsigned char *line, int x0, int x1)
 {
 	int a0, a1, b0, b1, a;
 
+	if (x1 <= x0)
+		return;
+
 	a0 = x0 >> 3;
 	a1 = x1 >> 3;
 
@@ -278,24 +311,25 @@ typedef struct fz_faxd_s fz_faxd;
 
 enum
 {
-	SNORMAL,	/* neutral state, waiting for any code */
-	SMAKEUP,	/* got a 1d makeup code, waiting for terminating code */
-	SEOL,		/* at eol, needs output buffer space */
-	SH1, SH2,	/* in H part 1 and 2 (both makeup and terminating codes) */
-	SDONE		/* all done */
+	STATE_NORMAL,	/* neutral state, waiting for any code */
+	STATE_MAKEUP,	/* got a 1d makeup code, waiting for terminating code */
+	STATE_EOL,		/* at eol, needs output buffer space */
+	STATE_H1, STATE_H2,	/* in H part 1 and 2 (both makeup and terminating codes) */
+	STATE_DONE		/* all done */
 };
 
 struct fz_faxd_s
 {
+	fz_context *ctx;
 	fz_stream *chain;
 
 	int k;
-	int endofline;
-	int encodedbytealign;
+	int end_of_line;
+	int encoded_byte_align;
 	int columns;
 	int rows;
-	int endofblock;
-	int blackis1;
+	int end_of_block;
+	int black_is_1;
 
 	int stride;
 	int ridx;
@@ -311,19 +345,18 @@ struct fz_faxd_s
 	unsigned char *rp, *wp;
 };
 
-static inline void
-eatbits(fz_faxd *fax, int nbits)
+static inline void eat_bits(fz_faxd *fax, int nbits)
 {
 	fax->word <<= nbits;
 	fax->bidx += nbits;
 }
 
-static inline int
-fillbits(fz_faxd *fax)
+static int
+fill_bits(fz_faxd *fax)
 {
 	while (fax->bidx >= 8)
 	{
-		int c = fz_readbyte(fax->chain);
+		int c = fz_read_byte(fax->chain);
 		if (c == EOF)
 			return EOF;
 		fax->bidx -= 8;
@@ -332,8 +365,8 @@ fillbits(fz_faxd *fax)
 	return 0;
 }
 
-static inline int
-getcode(fz_faxd *fax, const cfd_node *table, int initialbits)
+static int
+get_code(fz_faxd *fax, const cfd_node *table, int initialbits)
 {
 	unsigned int word = fax->word;
 	int tidx = word >> (32 - initialbits);
@@ -348,14 +381,14 @@ getcode(fz_faxd *fax, const cfd_node *table, int initialbits)
 		nbits = initialbits + table[tidx].nbits;
 	}
 
-	eatbits(fax, nbits);
+	eat_bits(fax, nbits);
 
 	return val;
 }
 
 /* decode one 1d code */
-static fz_error
-dec1d(fz_faxd *fax)
+static void
+dec1d(fz_context *ctx, fz_faxd *fax)
 {
 	int code;
 
@@ -363,18 +396,18 @@ dec1d(fz_faxd *fax)
 		fax->a = 0;
 
 	if (fax->c)
-		code = getcode(fax, cf_black_decode, cfd_black_initial_bits);
+		code = get_code(fax, cf_black_decode, cfd_black_initial_bits);
 	else
-		code = getcode(fax, cf_white_decode, cfd_white_initial_bits);
+		code = get_code(fax, cf_white_decode, cfd_white_initial_bits);
 
 	if (code == UNCOMPRESSED)
-		return fz_throw("uncompressed data in faxd");
+		fz_throw(ctx, "uncompressed data in faxd");
 
 	if (code < 0)
-		return fz_throw("negative code in 1d faxd");
+		fz_throw(ctx, "negative code in 1d faxd");
 
 	if (fax->a + code > fax->columns)
-		return fz_throw("overflow in 1d faxd");
+		fz_throw(ctx, "overflow in 1d faxd");
 
 	if (fax->c)
 		setbits(fax->dst, fax->a, fax->a + code);
@@ -384,38 +417,36 @@ dec1d(fz_faxd *fax)
 	if (code < 64)
 	{
 		fax->c = !fax->c;
-		fax->stage = SNORMAL;
+		fax->stage = STATE_NORMAL;
 	}
 	else
-		fax->stage = SMAKEUP;
-
-	return fz_okay;
+		fax->stage = STATE_MAKEUP;
 }
 
 /* decode one 2d code */
-static fz_error
-dec2d(fz_faxd *fax)
+static void
+dec2d(fz_context *ctx, fz_faxd *fax)
 {
 	int code, b1, b2;
 
-	if (fax->stage == SH1 || fax->stage == SH2)
+	if (fax->stage == STATE_H1 || fax->stage == STATE_H2)
 	{
 		if (fax->a == -1)
 			fax->a = 0;
 
 		if (fax->c)
-			code = getcode(fax, cf_black_decode, cfd_black_initial_bits);
+			code = get_code(fax, cf_black_decode, cfd_black_initial_bits);
 		else
-			code = getcode(fax, cf_white_decode, cfd_white_initial_bits);
+			code = get_code(fax, cf_white_decode, cfd_white_initial_bits);
 
 		if (code == UNCOMPRESSED)
-			return fz_throw("uncompressed data in faxd");
+			fz_throw(ctx, "uncompressed data in faxd");
 
 		if (code < 0)
-			return fz_throw("negative code in 2d faxd");
+			fz_throw(ctx, "negative code in 2d faxd");
 
 		if (fax->a + code > fax->columns)
-			return fz_throw("overflow in 2d faxd");
+			fz_throw(ctx, "overflow in 2d faxd");
 
 		if (fax->c)
 			setbits(fax->dst, fax->a, fax->a + code);
@@ -425,42 +456,42 @@ dec2d(fz_faxd *fax)
 		if (code < 64)
 		{
 			fax->c = !fax->c;
-			if (fax->stage == SH1)
-				fax->stage = SH2;
-			else if (fax->stage == SH2)
-				fax->stage = SNORMAL;
+			if (fax->stage == STATE_H1)
+				fax->stage = STATE_H2;
+			else if (fax->stage == STATE_H2)
+				fax->stage = STATE_NORMAL;
 		}
 
-		return fz_okay;
+		return;
 	}
 
-	code = getcode(fax, cf_2d_decode, cfd_2d_initial_bits);
+	code = get_code(fax, cf_2d_decode, cfd_2d_initial_bits);
 
 	switch (code)
 	{
 	case H:
-		fax->stage = SH1;
+		fax->stage = STATE_H1;
 		break;
 
 	case P:
-		b1 = findchangingcolor(fax->ref, fax->a, fax->columns, !fax->c);
+		b1 = find_changing_color(fax->ref, fax->a, fax->columns, !fax->c);
 		if (b1 >= fax->columns)
 			b2 = fax->columns;
 		else
-			b2 = findchanging(fax->ref, b1, fax->columns);
+			b2 = find_changing(fax->ref, b1, fax->columns);
 		if (fax->c) setbits(fax->dst, fax->a, b2);
 		fax->a = b2;
 		break;
 
 	case V0:
-		b1 = findchangingcolor(fax->ref, fax->a, fax->columns, !fax->c);
+		b1 = find_changing_color(fax->ref, fax->a, fax->columns, !fax->c);
 		if (fax->c) setbits(fax->dst, fax->a, b1);
 		fax->a = b1;
 		fax->c = !fax->c;
 		break;
 
 	case VR1:
-		b1 = 1 + findchangingcolor(fax->ref, fax->a, fax->columns, !fax->c);
+		b1 = 1 + find_changing_color(fax->ref, fax->a, fax->columns, !fax->c);
 		if (b1 >= fax->columns) b1 = fax->columns;
 		if (fax->c) setbits(fax->dst, fax->a, b1);
 		fax->a = b1;
@@ -468,7 +499,7 @@ dec2d(fz_faxd *fax)
 		break;
 
 	case VR2:
-		b1 = 2 + findchangingcolor(fax->ref, fax->a, fax->columns, !fax->c);
+		b1 = 2 + find_changing_color(fax->ref, fax->a, fax->columns, !fax->c);
 		if (b1 >= fax->columns) b1 = fax->columns;
 		if (fax->c) setbits(fax->dst, fax->a, b1);
 		fax->a = b1;
@@ -476,7 +507,7 @@ dec2d(fz_faxd *fax)
 		break;
 
 	case VR3:
-		b1 = 3 + findchangingcolor(fax->ref, fax->a, fax->columns, !fax->c);
+		b1 = 3 + find_changing_color(fax->ref, fax->a, fax->columns, !fax->c);
 		if (b1 >= fax->columns) b1 = fax->columns;
 		if (fax->c) setbits(fax->dst, fax->a, b1);
 		fax->a = b1;
@@ -484,7 +515,7 @@ dec2d(fz_faxd *fax)
 		break;
 
 	case VL1:
-		b1 = -1 + findchangingcolor(fax->ref, fax->a, fax->columns, !fax->c);
+		b1 = -1 + find_changing_color(fax->ref, fax->a, fax->columns, !fax->c);
 		if (b1 < 0) b1 = 0;
 		if (fax->c) setbits(fax->dst, fax->a, b1);
 		fax->a = b1;
@@ -492,7 +523,7 @@ dec2d(fz_faxd *fax)
 		break;
 
 	case VL2:
-		b1 = -2 + findchangingcolor(fax->ref, fax->a, fax->columns, !fax->c);
+		b1 = -2 + find_changing_color(fax->ref, fax->a, fax->columns, !fax->c);
 		if (b1 < 0) b1 = 0;
 		if (fax->c) setbits(fax->dst, fax->a, b1);
 		fax->a = b1;
@@ -500,7 +531,7 @@ dec2d(fz_faxd *fax)
 		break;
 
 	case VL3:
-		b1 = -3 + findchangingcolor(fax->ref, fax->a, fax->columns, !fax->c);
+		b1 = -3 + find_changing_color(fax->ref, fax->a, fax->columns, !fax->c);
 		if (b1 < 0) b1 = 0;
 		if (fax->c) setbits(fax->dst, fax->a, b1);
 		fax->a = b1;
@@ -508,36 +539,33 @@ dec2d(fz_faxd *fax)
 		break;
 
 	case UNCOMPRESSED:
-		return fz_throw("uncompressed data in faxd");
+		fz_throw(ctx, "uncompressed data in faxd");
 
 	case ERROR:
-		return fz_throw("invalid code in 2d faxd");
+		fz_throw(ctx, "invalid code in 2d faxd");
 
 	default:
-		return fz_throw("invalid code in 2d faxd (%d)", code);
+		fz_throw(ctx, "invalid code in 2d faxd (%d)", code);
 	}
-
-	return 0;
 }
 
 static int
-readfaxd(fz_stream *stm, unsigned char *buf, int len)
+read_faxd(fz_stream *stm, unsigned char *buf, int len)
 {
 	fz_faxd *fax = stm->state;
 	unsigned char *p = buf;
 	unsigned char *ep = buf + len;
 	unsigned char *tmp;
-	fz_error error;
 
-	if (fax->stage == SDONE)
+	if (fax->stage == STATE_DONE)
 		return 0;
 
-	if (fax->stage == SEOL)
+	if (fax->stage == STATE_EOL)
 		goto eol;
 
 loop:
 
-	if (fillbits(fax))
+	if (fill_bits(fax))
 	{
 		if (fax->bidx > 31)
 		{
@@ -549,13 +577,13 @@ loop:
 
 	if ((fax->word >> (32 - 12)) == 0)
 	{
-		eatbits(fax, 1);
+		eat_bits(fax, 1);
 		goto loop;
 	}
 
 	if ((fax->word >> (32 - 12)) == 1)
 	{
-		eatbits(fax, 12);
+		eat_bits(fax, 12);
 		fax->eolc ++;
 
 		if (fax->k > 0)
@@ -566,7 +594,7 @@ loop:
 				fax->dim = 1;
 			else
 				fax->dim = 2;
-			eatbits(fax, 1);
+			eat_bits(fax, 1);
 		}
 	}
 	else if (fax->k > 0 && fax->a == -1)
@@ -576,25 +604,21 @@ loop:
 			fax->dim = 1;
 		else
 			fax->dim = 2;
-		eatbits(fax, 1);
+		eat_bits(fax, 1);
 	}
 	else if (fax->dim == 1)
 	{
 		fax->eolc = 0;
-		error = dec1d(fax);
-		if (error)
-			return fz_rethrow(error, "cannot decode 1d code");
+		dec1d(stm->ctx, fax);
 	}
 	else if (fax->dim == 2)
 	{
 		fax->eolc = 0;
-		error = dec2d(fax);
-		if (error)
-			return fz_rethrow(error, "cannot decode 2d code");
+		dec2d(stm->ctx, fax);
 	}
 
 	/* no eol check after makeup codes nor in the middle of an H code */
-	if (fax->stage == SMAKEUP || fax->stage == SH1 || fax->stage == SH2)
+	if (fax->stage == STATE_MAKEUP || fax->stage == STATE_H1 || fax->stage == STATE_H2)
 		goto loop;
 
 	/* check for eol conditions */
@@ -609,9 +633,9 @@ loop:
 	goto loop;
 
 eol:
-	fax->stage = SEOL;
+	fax->stage = STATE_EOL;
 
-	if (fax->blackis1)
+	if (fax->black_is_1)
 	{
 		while (fax->rp < fax->wp && p < ep)
 			*p++ = *fax->rp++;
@@ -625,20 +649,20 @@ eol:
 	if (fax->rp < fax->wp)
 		return p - buf;
 
-	fax->rp = fax->dst;
-	fax->wp = fax->dst + fax->stride;
-
 	tmp = fax->ref;
 	fax->ref = fax->dst;
 	fax->dst = tmp;
 	memset(fax->dst, 0, fax->stride);
 
-	fax->stage = SNORMAL;
+	fax->rp = fax->dst;
+	fax->wp = fax->dst + fax->stride;
+
+	fax->stage = STATE_NORMAL;
 	fax->c = 0;
 	fax->a = -1;
 	fax->ridx ++;
 
-	if (!fax->endofblock && fax->rows)
+	if (!fax->end_of_block && fax->rows)
 	{
 		if (fax->ridx >= fax->rows)
 			goto rtc;
@@ -653,13 +677,13 @@ eol:
 			fax->dim = 2;
 	}
 
-	/* if endofline & encodedbytealign, EOLs are *not* optional */
-	if (fax->encodedbytealign)
+	/* if end_of_line & encoded_byte_align, EOLs are *not* optional */
+	if (fax->encoded_byte_align)
 	{
-		if (fax->endofline)
-			eatbits(fax, (12 - fax->bidx) & 7);
+		if (fax->end_of_line)
+			eat_bits(fax, (12 - fax->bidx) & 7);
 		else
-			eatbits(fax, (8 - fax->bidx) & 7);
+			eat_bits(fax, (8 - fax->bidx) & 7);
 	}
 
 	/* no more space in output, don't decode the next row yet */
@@ -669,86 +693,84 @@ eol:
 	goto loop;
 
 rtc:
-	fax->stage = SDONE;
+	fax->stage = STATE_DONE;
 	return p - buf;
 }
 
 static void
-closefaxd(fz_stream *stm)
+close_faxd(fz_context *ctx, void *state_)
 {
-	fz_faxd *fax = stm->state;
+	fz_faxd *fax = (fz_faxd *)state_;
 	int i;
 
 	/* if we read any extra bytes, try to put them back */
 	i = (32 - fax->bidx) / 8;
 	while (i--)
-		fz_unreadbyte(fax->chain);
+		fz_unread_byte(fax->chain);
 
 	fz_close(fax->chain);
-	fz_free(fax->ref);
-	fz_free(fax->dst);
-	fz_free(fax);
+	fz_free(ctx, fax->ref);
+	fz_free(ctx, fax->dst);
+	fz_free(ctx, fax);
 }
 
+/* Default: columns = 1728, end_of_block = 1, the rest = 0 */
 fz_stream *
-fz_openfaxd(fz_stream *chain, fz_obj *params)
+fz_open_faxd(fz_stream *chain,
+	int k, int end_of_line, int encoded_byte_align,
+	int columns, int rows, int end_of_block, int black_is_1)
 {
-	fz_faxd *fax;
-	fz_obj *obj;
+	fz_context *ctx = chain->ctx;
+	fz_faxd *fax = NULL;
 
-	fax = fz_malloc(sizeof(fz_faxd));
-	fax->chain = chain;
+	fz_var(fax);
 
-	fax->ref = nil;
-	fax->dst = nil;
+	fz_try(ctx)
+	{
+		fax = fz_malloc_struct(ctx, fz_faxd);
+		fax->chain = chain;
 
-	fax->k = 0;
-	fax->endofline = 0;
-	fax->encodedbytealign = 0;
-	fax->columns = 1728;
-	fax->rows = 0;
-	fax->endofblock = 1;
-	fax->blackis1 = 0;
+		fax->ref = NULL;
+		fax->dst = NULL;
 
-	obj = fz_dictgets(params, "K");
-	if (obj) fax->k = fz_toint(obj);
+		fax->k = k;
+		fax->end_of_line = end_of_line;
+		fax->encoded_byte_align = encoded_byte_align;
+		fax->columns = columns;
+		fax->rows = rows;
+		fax->end_of_block = end_of_block;
+		fax->black_is_1 = black_is_1;
 
-	obj = fz_dictgets(params, "EndOfLine");
-	if (obj) fax->endofline = fz_tobool(obj);
+		fax->stride = ((fax->columns - 1) >> 3) + 1;
+		fax->ridx = 0;
+		fax->bidx = 32;
+		fax->word = 0;
 
-	obj = fz_dictgets(params, "EncodedByteAlign");
-	if (obj) fax->encodedbytealign = fz_tobool(obj);
+		fax->stage = STATE_NORMAL;
+		fax->a = -1;
+		fax->c = 0;
+		fax->dim = fax->k < 0 ? 2 : 1;
+		fax->eolc = 0;
 
-	obj = fz_dictgets(params, "Columns");
-	if (obj) fax->columns = fz_toint(obj);
+		fax->ref = fz_malloc(ctx, fax->stride);
+		fax->dst = fz_malloc(ctx, fax->stride);
+		fax->rp = fax->dst;
+		fax->wp = fax->dst + fax->stride;
 
-	obj = fz_dictgets(params, "Rows");
-	if (obj) fax->rows = fz_toint(obj);
+		memset(fax->ref, 0, fax->stride);
+		memset(fax->dst, 0, fax->stride);
+	}
+	fz_catch(ctx)
+	{
+		if (fax)
+		{
+			fz_free(ctx, fax->dst);
+			fz_free(ctx, fax->ref);
+		}
+		fz_free(ctx, fax);
+		fz_close(chain);
+		fz_rethrow(ctx);
+	}
 
-	obj = fz_dictgets(params, "EndOfBlock");
-	if (obj) fax->endofblock = fz_tobool(obj);
-
-	obj = fz_dictgets(params, "BlackIs1");
-	if (obj) fax->blackis1 = fz_tobool(obj);
-
-	fax->stride = ((fax->columns - 1) >> 3) + 1;
-	fax->ridx = 0;
-	fax->bidx = 32;
-	fax->word = 0;
-
-	fax->stage = SNORMAL;
-	fax->a = -1;
-	fax->c = 0;
-	fax->dim = fax->k < 0 ? 2 : 1;
-	fax->eolc = 0;
-
-	fax->ref = fz_malloc(fax->stride);
-	fax->dst = fz_malloc(fax->stride);
-	fax->rp = fax->dst;
-	fax->wp = fax->dst + fax->stride;
-
-	memset(fax->ref, 0, fax->stride);
-	memset(fax->dst, 0, fax->stride);
-
-	return fz_newstream(fax, readfaxd, closefaxd);
+	return fz_new_stream(ctx, fax, read_faxd, close_faxd);
 }
